#import "MyOpenGLVideoView.h"

#define BailErr(x) {err = x; if(err != noErr) goto bail;}
#define BailIfNull(x) {gSeqGrab = x; if(gSeqGrab == nil) goto bail;}

static MyOpenGLVideoView	*myViewObject;	// our MyQuickDrawView object

@implementation MyOpenGLVideoView

//Initialize stuff at the beginning...
- (id) initWithCoder: (NSCoder *) coder
{
	self = [super initWithCoder: coder] ;
   
	NSOpenGLPixelFormatAttribute attrs[] = 
    {
        NSOpenGLPFADoubleBuffer, //Our Wishlist for the OpenGL Context
		NSOpenGLPFAAccelerated,
		NSOpenGLPFADepthSize, 16,
        nil
    } ;
	
    NSOpenGLPixelFormat* pixFmt ;
	
    pixFmt = [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs] ;
    [self setPixelFormat: pixFmt] ;

	//don't wait for vertical retrace to update image. (We have LCDs anyways...)
	long swapInterval = 0 ;
    [[self openGLContext]
        setValues: &swapInterval
	 forParameter: NSOpenGLCPSwapInterval ] ;	
	
	//some Stuff for interaction with Tree
	rotation = 0;
	angleX = 0;
	angleY = -30;
	texX = 256;
	texY = 128;
	
	saveOpenGLVideoViewObjectForCallback(self);
	
	//initialize and start all that QT-Stuff
	[ self doSeqGrab: [self bounds] ];
	
	// allocate a texture name to store the video from iSight
	glGenTextures (1, &videoTexture);
	// select our current texture. All following Tex-Commands modify that texture
	glBindTexture( GL_TEXTURE_2D, videoTexture );
	
	// when texture area is small, bilinear filter the original (usually use MIPMaps!)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	// when texture area is large, bilinear filter the original
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	// the texture wraps over at the edges (repeat)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	
	// select way to mix texture with color, we'll just add the colors...
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD );
	
	//build texture;
	glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE, texX, texY, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, workGL->imageData);	

    return self ;
}

//cleaning up all that mess.
- (void) dealloc
{
	[ self endGrab ];
}

// Perform clean-up when we are finished recording
-(void)endGrab
{
    ComponentResult result;
    OSErr 			err;
	
    // kill our sequence grabber idle timer first
    [gMyTimer invalidate];
    [gMyTimer release];
    
    // stop recording
    SGStop(gSeqGrab);
	
    // end our decompression sequences
    err = CDSequenceEnd(gDecomSeq);
	
    // finally, close our sequence grabber component
    result = CloseComponent(gSeqGrab);
    
    // get rid of our gworld
    DisposeGWorld(offScreen);
}


//We got a new image from iSight, plz update display.
- (void)updateImage:(char*)data
{
	// do somwthing with videoTexture
	glBindTexture( GL_TEXTURE_2D, videoTexture );
	
	glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE, texX, texY, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, data);	
	//the following should do the same, just a little faster, but produces ulgy artefacts.
	//glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, texX, texY, GL_LUMINANCE, GL_UNSIGNED_BYTE, data);	
	
	//let the tree rotate with each new image :-)
	rotation += 0.5;

	//plz update display if you find the time for that... (e.g. it is allready drawing, so let it finish)
	//will call drawRect if it has time to do so...
	[self setNeedsDisplay: YES] ;
}

//This (overloaded) method does the actual drawing. 
//It is called everytime the texture changes or the mouse is dragged...
- (void) drawRect: (NSRect) bounds
{
	//Do some setup for camera
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//Setup Frustum
	glViewport (0,0, [self bounds].size.width, [self bounds].size.height );
	gluPerspective(40, [self bounds].size.width / [self bounds].size.height ,.01,100);
	
	//Position the camera
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0,1,3.5, 0,0,0, 0,1,0);
	
	glRotatef (angleX, 0,1,0);
	glRotatef (-angleY, 1,0,0);
	
	//Make some ugly blue background
    glClearColor( 0, 0.5, 1, 1 ) ;
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
	
	// Stuff shall overlapp correctly!
	glEnable ( GL_DEPTH_TEST );
	
	//No Textures for line and "gras"
	glDisable ( GL_TEXTURE_2D );
	
	//get some green
	glColor3f (0.3,0.8,0.5);	
	
	//make a polygon
	glBegin (GL_QUADS );
	{ //note: brackets are only there to make code look nice! no function!
		glVertex3f (-1,-0.1,1);
		glVertex3f (1,-0.1,1);
		glVertex3f (1,-0.1,-1);
		glVertex3f (-1,-0.1,-1);
	}
	glEnd();
	
	//make some "brown" line
	glBegin (GL_LINES);
	{
		glColor3f (1,0.6,0.2);
		glVertex3f (0,-10,0);
		glVertex3f (0,10,0);
	}
	glEnd();
	
	//let the later stuff rotate
	glRotatef (rotation, 0,1,0);
	
	//we want the texture displayed
	glEnable( GL_TEXTURE_2D );
	
	//on a triangle
	glBegin( GL_TRIANGLES ) ;
	{       
		glColor3f( 1,0,0); glTexCoord2f(0,1); glVertex3f( -1.0f, 0.0f, 0.0f ) ;
		glColor3f( 0,1,0); glTexCoord2f(1,1); glVertex3f( 1.0f, 0.0f, 0.0f ) ;
		//If you Leave the range 0..1 for texture coordinates, the texture will be repeated
		glColor3f( 0,0,1); glTexCoord2f(0.5,-1); glVertex3f( 0.0f, 2.0f, 0.0f ) ;
	}
    glEnd() ;
		
	//done. As this is Double-Buffered, instruct system to draw at will. (long pipeline -> async!)
    [[self openGLContext] flushBuffer] ;
}

//Look someone klicked a MouseButton...
- (void) mouseDown: (NSEvent *) theEvent
{
	NSLog(@"Klick.\n");
	mousePoint = [theEvent locationInWindow] ;
	[self mouseDragged: theEvent] ;
}

//That Mouse moved... this is called independend form mouseDown! al 0.1 sec or something.
- (void) mouseDragged: (NSEvent *) theEvent
{
	NSPoint freshMousePoint = [theEvent locationInWindow] ;
	//adjust camera location
	angleX += ( freshMousePoint.x - mousePoint.x ) / 4.0;
	angleY += ( freshMousePoint.y - mousePoint.y ) / 4.0;
	
	//remember location to compute distance between two invocations
	mousePoint = freshMousePoint;

	// redraw
	[self setNeedsDisplay: YES] ;
}


- (IBAction)toggleSobel:(id)sender
{
	if (Sobel)
		Sobel = false;
	else
		Sobel = true;	
}

//As QT is a fscking old mess, we have to tess the SeqenceGraber: "I have CPU-Cycles to spare
//"plz do something." Nobody needed Multitasking anyway...
-(void) sgIdleTimer:(id)sender
{
    OSErr err;
	
    err = SGIdle(gSeqGrab);
    if (err != noErr)
    {
		NSLog(@"fucked up with timer!");
		// ...to fix this we simply call SGStop and SGStartRecord again
        // hopefully fixing any ...		
        SGStop(gSeqGrab);
        SGStartRecord(gSeqGrab);
    }
}

//Create Offscreen Buffer and start all that QT-Mess
-(OSErr) doSeqGrab:(NSRect)grabRect
{
    OSErr	err = noErr;
    
    // initialize the movie toolbox
    err = EnterMovies();
    BailErr(err);
    
    // open the sequence grabber component and initialize it
    gSeqGrab = OpenDefaultComponent(SeqGrabComponentType, 0);
    BailIfNull(gSeqGrab);
    
    err = SGInitialize(gSeqGrab);
    BailErr(err);
    
	// specify the destination data reference for a record operation
	// tell it we're not making a movie
	// if the flag seqGrabDontMakeMovie is used, the sequence grabber still calls
	// your data function, but does not write any data to the movie file
	// writeType will always be set to seqGrabWriteAppend
    err = SGSetDataRef(gSeqGrab, 0, 0, seqGrabDontMakeMovie);
    BailErr(err);
	
	
    // create a new sequence grabber video channel
    err = SGNewChannel(gSeqGrab, VideoMediaType, &gSGChanVideo);
    BailErr(err);
	
	//Tell preferred size (must be something nice! 640x480, 320x240, 176x144, all those pal, half-ntsc, ...
    gBoundsRect.top 	= 0;
    gBoundsRect.left 	= 0;
    gBoundsRect.bottom 	= 240;
    gBoundsRect.right 	= 320;
    err = SGSetChannelBounds(gSeqGrab, &gBoundsRect);
    
	//Allocate some temporary OpenCV Images.
	CVImageSize.width = 320;
    CVImageSize.height = 240;
	work = cvCreateImage (CVImageSize, IPL_DEPTH_16S, 1);
	work2 = cvCreateImage (CVImageSize, IPL_DEPTH_8U, 1);
	glSize.width = texX;
	glSize.height = texY;
	workGL = cvCreateImage (glSize, IPL_DEPTH_8U, 1); 
	
	
    // create the Offscreen GWorld
    err = QTNewGWorld(&offScreen,			// returned GWorld
    				  k32ARGBPixelFormat,		// pixel format (Note to self: k32RGBAPixelFormat?)
    				  &gBoundsRect,				// bounding rectangle
    				  0,						// color table
    				  NULL,						// graphic device handle
    				  0);						// flags
    BailErr(err);
	
    // lock the pixmap and make sure it's locked because
    // we can't decompress into an unlocked PixMap
    if(!LockPixels(GetPortPixMap(offScreen)))
    {
    	BailErr(-1);
    }
	
	//marry Offscreen and SeqGrabber
    err = SGSetGWorld(gSeqGrab, offScreen, GetMainDevice());
    BailErr(err);
	
    // set the bounds for the channel
    err = SGSetChannelBounds(gSGChanVideo, &gBoundsRect);
    BailErr(err);
    
    // set the usage for our new video channel to avoid playthrough
    // note: we do not set seqGrabPlayDuringRecord because if you set this flag
    // the data from the channel may be played during the record operation,
    // if the destination buffer is onscreen. However, playing the
    // data may affect the quality of the recorded sequence by causing frames 
    // to be dropped...something we definitely want to avoid
    err = SGSetChannelUsage(gSGChanVideo, seqGrabRecord);
    BailErr(err);
    
    // The following function is called every time, an image is ready. Here we will
	// e.g. Sobelfilter the image.
    err = SGSetDataProc(gSeqGrab,NewSGDataUPP(&mySGDataProc),NULL);
    BailErr(err);
	
    /* lights...camera... */
    err = SGPrepare(gSeqGrab,false,true);
    BailErr(err);
    
    // start recording!!
    err = SGStartRecord(gSeqGrab);
    BailErr(err);
	
	// setup a timer to idle the sequence grabber... 
    gMyTimer = [[NSTimer scheduledTimerWithTimeInterval:0.01		// interval, 0.1 seconds might suffice.
												 target:self
											   selector:@selector(sgIdleTimer:)		// call this method
											   userInfo:nil
												repeats:YES] retain];					// repeat until we cancel it

	////////////
    // Camera frames are now put into some buffer. Possibly in compressed form. Ne now need to extract those
	///////////
	
	ImageDescriptionHandle	imageDesc = (ImageDescriptionHandle)NewHandle(0);
    
    // retrieve a channels current sample description, the channel returns a sample description that is
    // appropriate to the type of data being captured. Hopefully identical to _request_ for resolution earlier
    err = SGGetChannelSampleDescription(gSGChanVideo,(Handle)imageDesc);
    BailErr(err);
	
	NSLog (@"Source-Size: x = %d, y = %d",  (**imageDesc).width,  (**imageDesc).height); 
		
    // begin the process of decompressing a sequence of frames
    // this is a set-up call and is only called once for the sequence - the ICM will interrogate different codecs
    // and construct a suitable decompression chain, as this is a time consuming process we don't want to do this
    // once per frame (eg. by using DecompressImage)
    // for more information see Ice Floe #8 http://developer.apple.com/quicktime/icefloe/dispatch008.html
    // the destination is specified as the GWorld
    err = DecompressSequenceBegin(&gDecomSeq,	// pointer to field to receive unique ID for sequence
								  imageDesc,		// handle to image description structure
								  offScreen,		// port for the DESTINATION image
								  NULL,			// graphics device handle, if port is set, set to NULL
								  NULL,			// source rectangle defining the portion of the image to decompress
								  NULL,	// transformation matrix
								  srcCopy,		// transfer mode specifier
								  NULL,			// clipping region in dest. coordinate system to use as a mask
								  0,						// flags
								  codecNormalQuality,		// accuracy in decompression
								  bestSpeedCodec);		// compressor identifier or special identifiers ie. bestSpeedCodec
    BailErr(err);
    
    DisposeHandle((Handle)imageDesc);
    imageDesc = NULL;
    
	BailErr(err);
	
bail:
		
		return err;
}

-(GWorldPtr)gworld
{
    return offScreen;
}

-(ImageSequence)decomSeq
{
    return gDecomSeq;
}

-(SGChannel)sgChanVideo
{
    return gSGChanVideo;
}

-(Rect)boundsRect
{
    return gBoundsRect;
}

-(bool)findEdges
{
	return Sobel;
}

-(IplImage *)work
{
	return work;
}

-(IplImage *)work2
{
	return work2;
}

-(IplImage *)workGL
{
	return workGL;
}


@end



/* ---------------------------------------------------------------------- */
/* sequence grabber data procedure - this is where the work is done       */
/* ---------------------------------------------------------------------- */
/* mySGDataProc - the sequence grabber calls the data function whenever
any of the grabbers channels write digitized data to the destination movie file.

NOTE: We really mean any, if you have an audio and video channel then the DataProc will
be called for either channel whenever data has been captured. Be sure to check which
channel is being passed in. In this example we never create an audio channel so we know
we're always dealing with video.

This data function decompresses captured video data into an offscreen GWorld,
then transfers the frame to an onscreen window.

For more information refer to Inside Macintosh: QuickTime Components, page 5-120
c - the channel component that is writing the digitized data.
p - a pointer to the digitized data.
len - the number of bytes of digitized data.
offset - a pointer to a field that may specify where you are to write the digitized data,
and that is to receive a value indicating where you wrote the data.
chRefCon - per channel reference constant specified using SGSetChannelRefCon.
time	- the starting time of the data, in the channels time scale.
writeType - the type of write operation being performed.
seqGrabWriteAppend - Append new data.
seqGrabWriteReserve - Do not write data. Instead, reserve space for the amount of data
specified in the len parameter.
seqGrabWriteFill - Write data into the location specified by offset. Used to fill the space
previously reserved with seqGrabWriteReserve. The Sequence Grabber may
call the DataProc several times to fill a single reserved location.
refCon - the reference constant you specified when you assigned your data function to the sequence grabber.
*/


pascal OSErr mySGDataProc(SGChannel c, 
						  Ptr p,
						  long len,
						  long *offset,
						  long chRefCon,
						  TimeValue time,
						  short writeType, 
						  long refCon)
{
#pragma unused(offset,chRefCon,time,writeType)
    
    CodecFlags 		ignore;
	ComponentResult err = noErr;
	
    if([myViewObject gworld]) 
    {			
        // decompress a frame into the GWorld - can queue a frame for async decompression when passed in a completion proc
        // once the image is in the GWorld it can be manipulated at will
        err = DecompressSequenceFrameS([myViewObject decomSeq],	// sequence ID returned by DecompressSequenceBegin
									   p,						// pointer to compressed image data
									   len,					// size of the buffer
									   0,						// in flags
									   &ignore,				// out flags
									   NULL);					// async completion proc
        BailErr(err);
		
        // ******  IMAGE IS NOW IN THE GWORLD ****** //
	}
	
	// Transform Pixmap for openCV...
	PixMapHandle	pix = GetGWorldPixMap([myViewObject gworld]);
    Rect		bounds;
	
	GetPortBounds([myViewObject gworld], &bounds);
    OffsetRect(&bounds, -bounds.left, -bounds.top);

	int x,y,i = 0;
	UInt32* baseAddr;
	
    for (y = 0; y < bounds.bottom; ++y)
    {
        baseAddr = (UInt32*)(GetPixBaseAddr(pix) + y * GetPixRowBytes(pix));
        for (x = 0; x < bounds.right; x++)
        {
            UInt32			color;
            unsigned long			R;
            unsigned long			G;
            unsigned long			B;
			
            // get the RGB of the color
            color = *baseAddr;
            R = (color & 0x00FF0000) >> 16;
            G = (color & 0x0000FF00) >> 8;
            B = (color & 0x000000FF) >> 0;
			
			[myViewObject work2]->imageData[ i ] = (R+G+B) / 3; //grayscale is nice so far...
            baseAddr++;
			i++;
        }
    }
	
	if ([ myViewObject findEdges ]){
		
		//Run the Filter. Result is written into work.
		cvSobel( [myViewObject work2],[myViewObject work], 1, 1, 5);
		
		//Map to usigned 8 Bit, Write result to work2.
		cvConvertScale ([myViewObject work], [myViewObject work2], 1.0, 0.0);
	}
	cvResize ( [myViewObject work2], [myViewObject workGL], CV_INTER_LINEAR);
	
    [ myViewObject updateImage: [myViewObject workGL]->imageData];
	
bail:
		
		return err;
}

//////////
//
// saveQDViewObjectForCallback
//
// This routine stores a reference to our MyQuickDrawView object. We'll
// need this so we can call into methods in this class from outside the
// implementation of the class methods (specifically, from our SGDataProc
// C routine above).
//
//////////

void saveOpenGLVideoViewObjectForCallback(void *theObject)
{
	myViewObject = (MyOpenGLVideoView *)theObject;
}


